package org.fluff.splodge.scim;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;

import org.fluff.splodge.util.*;

import java.lang.reflect.Proxy;

/**
A SCIMConnection is the gateway to a peer SCIMConnection on a remote machine.

22nd June 2004, 14:24 -> 14:26; 14:54 ->16:02
23rd June 2004, 09:15 -> 09:18

@author Phil Norman
*/

public class SCIMConnection implements Runnable {
	private static final byte SERVER_REFERENCE = 1;
	private static final byte METHOD_CALL = 2;
	private static final byte METHOD_RESULT = 3;
	private static final byte REFERENCE_LOST = 4;
	private static final byte METHOD_EXCEPTION = 5;

	private Socket socket;
	private ObjectOutputStream out;
	private ObjectInputStream in;
	
	private SCIMRemote serverRef;
	
	private Map resultPipes = Collections.synchronizedMap(new HashMap());
	private IdGenerator requestIdGenerator = new IdGenerator();
	
	private Map references = Collections.synchronizedMap(new HashMap());
	private IdGenerator referenceIdGenerator = new IdGenerator();
	
	private static ThreadLocal threadConnection = new ThreadLocal();
	
	public SCIMConnection(Socket socket) throws IOException {
		// Note - always open the output stream before the input stream, otherwise both ends
		// will block forever.  This is because opening the object input stream causes it to check for
		// the correct header.
		this.socket = socket;
		out = new ObjectOutputStream(socket.getOutputStream());
		in = new ObjectInputStream(socket.getInputStream());
		start();
	}
	
	private void start() {
		(new Thread(this, "SCIM Connection Handler")).start();
	}
	
	public void run() {
		threadConnection.set(this);
		try {
			while (true) {
				try {
					int code = in.readByte();
					switch (code) {
						case SERVER_REFERENCE:
							readServerReference();
							break;
						case METHOD_CALL:
							readMethodCall();
							break;
						case METHOD_RESULT:
							readMethodResult();
							break;
						case REFERENCE_LOST:
							readReferenceLost();
							break;
						 case METHOD_EXCEPTION:
						 	readMethodException();
						 	break;
						 default:
						 	System.err.println("Warning: received unknown code " + code);
					}
				} catch (Exception ex) {
					if (ex instanceof IOException) {
						throw (IOException) ex;
					}
					ex.printStackTrace();
				}
			}
		} catch (IOException ex) { }
		connectionClosed();
	}
	
	public static SCIMConnection getConnectionForThread() {
		return (SCIMConnection) threadConnection.get();
	}
	
	public int registerAsRemotelyReferenced(Object remotelyReferencedObject) {
		int referenceId = referenceIdGenerator.get();
		references.put(new Integer(referenceId), remotelyReferencedObject);
		return referenceId;
	}
	
	public Object getProxy(Class remoteClass, int referenceId) {
		Class[] interfaces = getRemoteInterfaces(remoteClass);
		InvocationHandler handler = new RemoteInvocationHandler(referenceId);
		Object result = Proxy.newProxyInstance(remoteClass.getClassLoader(), interfaces, handler);
		ReferenceNotifier.getInstance().addListener(result, new ReferenceLossNotifier(referenceId));
		return result;
	}
	
	private Class[] getRemoteInterfaces(Class cls) {
		HashSet results = new HashSet();
		while (cls.equals(Object.class) == false) {
			Class[] inter = cls.getInterfaces();
			for (int i = 0; i < inter.length; i++) {
				if (SCIMRemote.class.isAssignableFrom(inter[i])) {
					results.add(inter[i]);
				}
			}
			cls = cls.getSuperclass();
		}
		return (Class[]) results.toArray(new Class[results.size()]);
	}
	
	public class ReferenceLossNotifier implements ReferenceListener {
		private int referenceId;
		
		public ReferenceLossNotifier(int referenceId) {
			this.referenceId = referenceId;
		}
		
		public void objectDestroyed() {
			try {
				sendReferenceLost(referenceId);
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}
	
	public class RemoteInvocationHandler implements InvocationHandler {
		private int referenceId;
		
		public RemoteInvocationHandler(int referenceId) {
			this.referenceId = referenceId;
		}

		public Object invoke(Object proxy, Method method, Object[] args) {
			return callRemoteMethod(referenceId, method.getName(), method.getParameterTypes(), args);
		}
	}
	
	private void connectionClosed() {
		Object[] keys = resultPipes.keySet().toArray();
		RuntimeException closedException = new RuntimeException("Connection closed.");
		closedException.fillInStackTrace();
		for (int i = 0; i < keys.length; i++) {
			BlockingPipe pipe = (BlockingPipe) resultPipes.get(keys[i]);
			pipe.add(new Result(closedException, true));
		}
	}
	
	public Object callRemoteMethod(int referenceId, String methodName, Class[] argTypes, Object[] args) {
		threadConnection.set(this);
		BlockingPipe resultPipe = new BlockingPipe();
		int reqId = requestIdGenerator.get();
		resultPipes.put(new Integer(reqId), resultPipe);
		try {
			synchronized (out) {
				out.writeByte(METHOD_CALL);
				out.writeInt(reqId);
				out.writeInt(referenceId);
				out.writeObject(methodName);
				out.writeObject(argTypes);
				out.writeObject(args);
			}
		} catch (IOException ex) {
			throw new RuntimeRemoteException("Failed to send method call request.", ex);
		}
		Result result = (Result) resultPipe.get();
		resultPipes.remove(new Integer(reqId));
		threadConnection.set(null);  // Just to make any bugs more obvious.
		if (result.isException()) {
			throw new RuntimeRemoteException("Exception in remote method call.", (Throwable) result.getResult());
		} else {
			return result.getResult();
		}
	}
	
	public synchronized SCIMRemote getServerReference() {
		while (serverRef == null) {
			try {
				wait();
			} catch (InterruptedException ex) {
				ex.printStackTrace();
			}
		}
		return serverRef;
	}
	
	public void sendReferenceLost(int referenceId) throws IOException {
		synchronized (out) {
			out.writeByte(REFERENCE_LOST);
			out.writeInt(referenceId);
		}
	}
	
	public void sendServerReference(SCIMRemote ref) throws IOException {
		threadConnection.set(this);
		synchronized (out) {
			out.writeByte(SERVER_REFERENCE);
			out.writeObject(ref);
		}
		threadConnection.set(null);
	}
	
	public void sendResult(int requestId, Object result, boolean isException) throws IOException {
		threadConnection.set(this);
		synchronized (out) {
			out.writeByte(isException ? METHOD_EXCEPTION : METHOD_RESULT);
			out.writeInt(requestId);
			out.writeObject(result);
		}
		threadConnection.set(null);
	}
	
	private synchronized void readServerReference() throws IOException, ClassNotFoundException {
		serverRef = (SCIMRemote) in.readObject();
		notifyAll();
	}
	
	private void readMethodCall() throws IOException, ClassNotFoundException {
		int reqId = in.readInt();
		int referenceId = in.readInt();
		String methodName = (String) in.readObject();
		Class[] argTypes = (Class[]) in.readObject();
		Object[] args = (Object[]) in.readObject();
		Object reference = getReference(referenceId);
		MethodCaller caller = new MethodCaller(this, reqId, reference, methodName, argTypes, args);
		caller.start();
	}
	
	private void readMethodResult() throws IOException, ClassNotFoundException {
		int reqId = in.readInt();
		Object result = in.readObject();
		getResultPipe(reqId).add(new Result(result, false));
	}
	
	private void readReferenceLost() throws IOException, ClassNotFoundException {
		int refId = in.readInt();
		references.remove(new Integer(refId));
	}
	
	private void readMethodException() throws IOException, ClassNotFoundException {
		int reqId = in.readInt();
		Throwable exception = (Throwable) in.readObject();
		getResultPipe(reqId).add(new Result(exception, true));
	}
	
	private Object getReference(int referenceId) {
		return references.get(new Integer(referenceId));
	}
	
	private BlockingPipe getResultPipe(int requestId) {
		return (BlockingPipe) resultPipes.get(new Integer(requestId));
	}
	
	private static class Result {
		private Object result;
		private boolean isException;
		
		public Result(Object result, boolean isException) {
			this.result = result;
			this.isException = isException;
		}
		
		public boolean isException() {
			return isException;
		}
		
		public Object getResult() {
			return result;
		}
	}
}
